## Class: Battery cell

# Packages
import numpy as np


# System variables
TIMESTEP = 1/(60*2) # [h]


# List of battery SOC and terminal voltage values
SOC_list = [0.0, 0.05, 0.10, 0.15, 0.20, 0.25, 0.30, 0.35, 0.40, 0.45, 0.50, 0.55, 0.60, 0.65, 0.70, 0.75, 0.80, 0.85, 0.90, 0.95, 1.00]
Voc_list = [2.03,3.08,3.2,3.22,3.25,3.265,3.28,3.29,3.295,3.3,3.305,3.310,3.315,3.323,3.33,3.34,3.35,3.358,3.365,3.375,3.6]


# Degradation parameters
z   = 0.57 # Ah throughput dependence
eta = 152.5 # C-rate dependence
Rg  = 8.3144621 # Universal gas constant
Ea  = 31500 # Activation energy
EoL = 20 # [%] End of life condition


def get_severity_factor_func(SOC, Ic, T):

	alpha = 2795.55
	beta  = 6716.7

	return (alpha*SOC + beta)*np.exp((-Ea + eta*Ic)/(Rg*(273.15 + T)))


def get_nominal_battery_life():
	Ic_nom  = 2.5
	SOC_nom = 0.35
	T_nom   = 25
	sigma = get_severity_factor_func(SOC_nom, Ic_nom, T_nom)

	return (EoL / sigma)**(1/z)


def get_battery_life(SOC, Ic, T):
	sigma = get_severity_factor_func(SOC, Ic, T)
	# print(SOC, Ic, T, sigma)
	return (EoL / sigma)**(1/z)


def get_severity_factor(SOC, Ic, T):
	return get_nominal_battery_life() / get_battery_life(SOC, Ic, T)


def get_nominal_calendar_aging():
	Ic_nom  = 0
	SOC_nom = 0.35
	T_nom   = 25
	sigma = get_severity_factor_func(SOC_nom, Ic_nom, T_nom)

	return (EoL / sigma)**(1/z)

def get_relative_calendar_aging(SOC, T):
	sigma = get_severity_factor_func(SOC, 0, T)
	return (EoL / sigma)**(1/z)


def get_calendar_aging(SOC, T):
	return (get_nominal_calendar_aging() / get_relative_calendar_aging(SOC, T))*10**(-1)


def get_aging(I, Ic, SOC, T):
	return get_severity_factor(SOC, Ic, T)*np.absolute(I)


class Cell:
	# Class variables
	# Cell: Lithium-ion LiFePO4 ANR26650
	V_nom = 3.2 # [V]
	C_nom = 2.5 # [Ah]
	P_nom = 25 # [W] OBS
	R_0   = 12*10**(-3) # [Ohm]

	# Instance variables
	def __init__(self, SOC0):
		# cell specs
		self.capacity    = Cell.C_nom
		self.rated_power = 25
		self.SOC		 = SOC0
		self.SOH         = 1
		self.temperature = 25
		self.R 			 = Cell.R_0

		# Cell utilization
		self.I           = 0
		self.Ic			 = 0 # C-rate
		self.Voc	     = np.interp(self.SOC, SOC_list, Voc_list)
		self.power       = 0
		self.Ah_ctrl     = 0 # Controllable Ah - throughput
		self.Ah_eff      = 0 # Effective Ah - throughput
		self.Ah_eff_ctrl = 0 # Effective controllable Ah - throughput
		self.Ah_max      = get_nominal_battery_life()


	def set_cell_current(self, power):
		self.I  = (self.Voc - np.sqrt(self.Voc**2 - 4*Cell.R_0*np.absolute(power)))/(2*Cell.R_0)
		if power < 0:
			self.I = self.I*(-1)

		self.Ic = np.absolute(self.I) / self.capacity
		return None


	def get_cell_current(self, power):
		I  = (self.Voc - np.sqrt(self.Voc**2 - 4*Cell.R_0*np.absolute(power)))/(2*Cell.R_0)
		if power < 0:
			I = I*(-1)
		return I


	def get_cell_c_rate(self, current):
		return np.absolute(current) / self.capacity


	def update_SOC(self):
		if self.capacity > 0:
			self.SOC = self.SOC - (self.I*TIMESTEP)/self.capacity
		return None


	def update_cell_Voc(self):
		self.Voc = np.interp(self.SOC, SOC_list, Voc_list)
		return None


	def update_controllable_Ah_throughput(self):
		self.Ah_ctrl = self.Ah_ctrl + np.absolute(self.I)*TIMESTEP
		return None


	def update_effective_Ah_throughput(self):
		self.Ah_eff = self.Ah_eff + get_aging(self.I, self.Ic, self.SOC, self.temperature)*TIMESTEP
		return None


	def update_effective_controllable_Ah_throughput(self):
		self.Ah_eff_ctrl = self.Ah_eff_ctrl + get_aging(self.I, self.Ic, self.SOC, self.temperature)*TIMESTEP


	def update_SOH(self):
		self.SOH = 1 - self.Ah_eff/self.Ah_max
		return None


	def update_cell(self, power):
		self.power = power
		self.set_cell_current(self.power)
		self.update_SOC()
		self.update_cell_Voc()
		self.update_effective_Ah_throughput()
		self.update_SOH()
		return None
