## Class: Ferry Controller

import sys
sys.path.insert(0, 'C:\\Users\\noldu\\OneDrive\\NTNU\\Master thesis\\Application\\ELCost')
sys.path.insert(0, 'C:\\Users\\noldu\\OneDrive\\NTNU\\Master thesis\\Application\\External')
sys.path.insert(0, 'C:\\Users\\noldu\\OneDrive\\NTNU\\Master thesis\\Application\\DB')


# Import modules
import elcost as EL
import pcu
import wind
import current as curr
import db


# Import packages
import numpy  as np
from datetime import datetime as dt
from datetime import timedelta as td
import pandas as pd


# Define variables
kW = 10**3
TIMESTEP = 1/(2*60) # [h]

# Get weather data
weather_db = "C:\\Users\\noldu\\OneDrive\\NTNU\\Master thesis\\Application\\DB\\weatherdatadb.db"
weather_data = db.query_all_db_data(weather_db, 'WeatherData')

# Get ocean current data
ocean_db = "C:\\Users\\noldu\\OneDrive\\NTNU\\Master thesis\\Application\\DB\\oceandatadb.db"
ocean_data = db.query_all_db_data(ocean_db, 'OceanData')

class FerryController:

	# Instance variable
	def __init__(self):
		# Ferry direction
		self.direction = 0

		# Ferry powers
		self.battery_power    = 0
		self.gen_set_power    = 0
		self.grid_power       = 0 # Supplies ferry load
		self.propulsion_power = 0
		self.hotel_load       = 0
		self.total_load       = 0
		self.consumption      = 0
		
		# Extras
		self.C_loss           = 0
		self.pcu_load         = 0
		self.wind_load        = 0
		self.current_load     = 0
		

	def set_power(self, model, profile, index):
		# Set weekend flags
		if index == 2312:
			EL.set_weekend_flag()
			pcu.set_weekend_flag()

		if profile.at[index, 'mode'] == 'Nightlay':
			input_power = model.ferry.power
			self.hotel_load    = profile.at[index, 'hotelPower']*kW
			self.total_load    = self.hotel_load
			self.gen_set_power = profile.at[index, 'genSetPower']*kW

			self.propulsion_power = 0
			self.battery_power    = (input_power + self.gen_set_power - self.total_load)*(-1)
			self.grid_power       = input_power + self.battery_power

		elif profile.at[index, 'mode'] == 'Transit':
			self.gen_set_power    = profile.at[index, 'genSetPower']*kW
			self.propulsion_power = profile.at[index, 'propulsionPower']*kW
			self.hotel_load       = profile.at[index, 'hotelPower']*kW

			FerryController.set_ferry_wind_load(self, index)
			FerryController.set_ferry_ocean_current_load(self, index)

			# Calculate total load
			self.total_load    = self.propulsion_power + self.hotel_load

			# Adjust for PCU load
			self.total_load    = self.total_load/(1 - self.pcu_load)

			# Add external load
			self.total_load    = self.total_load + self.wind_load + self.current_load

			# Calculate for ferry power loss
			total_power        = self.total_load*(1/(model.ferry.eff_board*model.ferry.eff_trafo*model.ferry.eff_conv))

			self.consumption   = self.consumption + total_power*(1/(2*60))
			self.battery_power = (total_power - self.gen_set_power)
			self.grid_power    = total_power

		elif profile.at[index, 'mode'] == 'Conn':
			self.gen_set_power = profile.at[index, 'genSetPower']*kW
			self.hotel_load    = profile.at[index, 'hotelPower']*kW
			self.total_load    = self.hotel_load

			total_power = self.hotel_load*(1/(model.ferry.eff_board*model.ferry.eff_trafo*model.ferry.eff_conv))

			self.consumption = 0

			self.propulsion_power = 0
			self.battery_power    = total_power - self.gen_set_power
			self.grid_power       = self.gen_set_power - self.battery_power			

		elif profile.at[index, 'mode'] == 'Discon':
			self.gen_set_power = profile.at[index, 'genSetPower']*kW
			self.hotel_load    = profile.at[index, 'hotelPower']*kW
			FerryController.set_ferry_pcu_load(self, profile, index)
			FerryController.set_ferry_direction(self, profile, index)

			total_power = self.hotel_load*(1/(model.ferry.eff_board*model.ferry.eff_trafo*model.ferry.eff_conv))

			self.propulsion_power = 0
			self.battery_power    = total_power - self.gen_set_power
			self.grid_power       = self.gen_set_power - self.battery_power

		elif profile.at[index, 'mode'] == 'Docked':
			input_power           = model.ferry.power
			self.propulsion_power = 0
			self.total_load       = profile.at[index, 'hotelPower']*kW
			self.gen_set_power    = profile.at[index, 'genSetPower']*kW
			self.battery_power    = (input_power + self.gen_set_power - profile.at[index, 'hotelPower']*kW)*(-1)
			self.grid_power       = input_power + self.battery_power

		return None


	def set_ferry_direction(self, profile, index):
		leg = profile.at[index, 'leg']
		if leg == 1:
			self.direction = 2
		else:
			self.direction = 1


	def set_ferry_pcu_load(self, profile, index):
		PCUmean = 25
		load_factor = 5

		t    = dt.strptime(profile.at[index, 'datetime'], '%Y-%m-%d %H:%M:%S') + td(seconds=30)
		t    = t.hour + t.minute/60 + t.second/(60*60)
		np.random.seed(index)
		PCUs = pcu.get_random_pcus(t)

		self.pcu_load = ((PCUs - PCUmean)/PCUmean) * load_factor * 0.01
		return None


	def set_ferry_wind_load(self, index):
		weather_index = int(index/(2*60))
		interp_index  = (index/(2*60)) - weather_index

		# Calculate interpolated wind speed
		wind_speed0   = weather_data.at[weather_index, 'windSpeed']
		wind_speed1   = weather_data.at[weather_index + 1, 'windSpeed']
		wind_speed    = wind_speed0 + (interp_index - 0)*((wind_speed1 - wind_speed0)/(1 - 0))
		# Calculate random wind speed
		np.random.seed(index)
		wind_speed = np.absolute(np.random.normal(wind_speed, 2))

		# Calculate interpolated wind direction
		wind_dir0   = weather_data.at[weather_index, 'windDir']
		wind_dir1   = weather_data.at[weather_index+1, 'windDir']
		wind_dir    = wind_dir0 + (interp_index - 0)*((wind_dir1 - wind_dir0)/(1 - 0))
		# Calculate random wind dir
		np.random.seed(index)
		wind_dir = np.absolute(np.random.normal(wind_dir, 10))

		P_Fw, P_Lw, Pw  = wind.calculate_wind_power(wind_speed, wind_dir)

		# Effect with regards to direction
		if self.direction == 2:
			self.wind_load = P_Fw
		else:
			self.wind_load = P_Fw*(-1)
		return None


	def set_ferry_ocean_current_load(self, index):
		current_index = int(index/(2*60))
		interp_index  = (index/(2*60)) - current_index

		# Calculate interpolated wind speed
		current_speed0   = ocean_data.at[current_index, 'seaCurrentSpeed']
		current_speed1   = ocean_data.at[current_index+1, 'seaCurrentSpeed']
		current_speed    = current_speed0 + (interp_index - 0)*((current_speed1 - current_speed0)/(1 - 0))
		# Calculate random current speed
		np.random.seed(index)
		current_speed = np.absolute(np.random.normal(current_speed, 0.2))

		# Calculate interpolated wind direction
		current_dir0   = ocean_data.at[current_index, 'seaCurrentDirection']
		current_dir1   = ocean_data.at[current_index+1, 'seaCurrentDirection']
		current_dir    = current_dir0 + (interp_index - 0)*((current_dir1 - current_dir0)/(1 - 0))
		# Calculate random current dir
		np.random.seed(index)
		current_dir = np.absolute(np.random.normal(current_dir, 10))

		P_Fc, P_Lc, Pc  = curr.calculate_current_power(current_speed, current_dir)

		# Effect with regards to direction
		if self.direction == 2:
			self.current_load = P_Fc
		else:
			self.current_load = P_Fc*(-1)
		return None


	def update_power_loss_cost(self, ferry, t):
		t = t.hour + t.minute/60 + t.second/(60*60)
		P_loss           = (1 - ferry.battery.get_efficiency(self.battery_power))*self.battery_power
		electricity_cost = EL.get_price(t)*np.absolute(P_loss)*10**(-3)*TIMESTEP
		tariff_cost      = EL.get_tariff_cost_sim(np.absolute(P_loss))

		self.C_loss = self.C_loss + electricity_cost + tariff_cost
		return None