## Battery info tab

import dash
import dash_core_components as dcc
import dash_html_components as html
import pandas as pd
import numpy as np
import plotly.graph_objs as go

from app import app

# Parameters
z   = 0.57 # Ah throughput dependence
eta = 152.5 # C-rate dependence
Rg  = 8.3144621 # Universal gas constant
Ea  = 31500 # Activation energy
EoL = 20 # [%] End of life condition

temp_list = [5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60]


def get_severity_factor(SOC, Ic, T):
	if SOC < 0.45:
		alpha = 2896.6
		beta  = 7411.2
	else:
		alpha = 2694.5
		beta  = 6022.2

	return (alpha*SOC + beta)*np.exp((-Ea + eta*Ic)/(Rg*(273.15 + T)))


def get_max_battery_life():
	Ic_nom  = 2.5
	SOC_nom = 0.45
	T_nom   = 25
	sigma = get_severity_factor(SOC_nom, Ic_nom, T_nom)

	return (EoL / sigma)**(1/z)


def get_battery_life(SOC, Ic, T):
	sigma = get_severity_factor(SOC, Ic, T)

	return (EoL / sigma)**(1/z)


def get_severity_factor_map(SOC, Ic, T):
	return get_max_battery_life() / get_battery_life(SOC, Ic, T)


layout = html.Div([
    dcc.Graph(id='severity-factor-map-with-slider'),
    dcc.Slider(
        id='temp-slider',
        min=temp_list[0],
        max=temp_list[len(temp_list)-1],
        value=25,
        marks={str(temp): str(temp) for temp in temp_list}
    )
])


@app.callback(
    dash.dependencies.Output('severity-factor-map-with-slider', 'figure'),
    [dash.dependencies.Input('temp-slider', 'value')])
def update_figure(selected_temp):
	SOC_list = np.linspace(0.20, 0.80, 60)
	Ic_list  = np.linspace(0, 25, 25)

	sigma_map_matrix = [[0]*len(Ic_list) for i in range(len(SOC_list))]

	for i in range(0, len(SOC_list)-1):
		for j in range(0, len(Ic_list)-1):
			sigma_map_matrix[i][j] = get_severity_factor_map(SOC_list[i], Ic_list[j], selected_temp)

	surface0 = go.Surface(
		y = SOC_list,
		x =  Ic_list,
		z = sigma_map_matrix
	)

	return {'data': [surface0]}