## Ocean current effect on ferry

# Modules
import numpy as np
import matplotlib.pyplot as plt


def calculate_current_power(current_speed, current_dir):
	# Ferry parameters
	ferry_length    = 84
	ferry_breadth   = 16
	ferry_draft     = 3.85

	# current conditions
	Vc        = current_speed
	gamma_c   = np.radians(current_dir)

	# current coefficients
	c_x = 0.45
	c_y = 0.65

	# Physical parameters
	rho = 997

	# Projected areas
	A_Fc = ferry_breadth*ferry_draft
	A_Lc = ferry_length*ferry_draft

	# current speed components
	u_c = Vc*np.cos(gamma_c)
	v_c = Vc*np.sin(gamma_c)

	# current coefficient approx components
	Cx = c_x*np.cos(gamma_c)*np.absolute(np.cos(gamma_c))
	Cy = c_y*np.sin(gamma_c)*np.sin(gamma_c)

	F_Fc = 0.5 * rho * (Vc**2) * Cx * A_Fc
	F_Lc = 0.5 * rho * (Vc**2) * Cy * A_Lc

	Fc = np.sqrt(F_Fc**2 + F_Lc**2)

	P_Fc = F_Fc*Vc
	P_Lc = F_Lc*Vc

	Pc = np.sqrt(P_Fc**2 + P_Lc**2)
	return P_Fc, P_Lc, Pc


def plot_current_power():
	speeds = np.linspace(0, 2, 100)
	current_dirs = np.linspace(0, 360, 8)

	for current_dir in current_dirs:
		Pc = []
		for speed in speeds:
			P = calculate_current_power(speed, current_dir)
			Pc.append(P*10**(-3))

		plt.plot(speeds, Pc, label='Direction [°]: %d'%current_dir)

	plt.grid(True)
	plt.xlabel('Ocean current speed [m/s]', fontsize='16')
	plt.ylabel('Ocean current power [kW]', fontsize='16')
	plt.legend()
	plt.title('Ocean current power in surge component', fontsize='16')
	plt.show()


def plot_current_coeffs():
	# current coefficients
	c_x = 0.45
	c_y = 0.65

	y1 = []
	y2 = []
	dirs = np.linspace(0, 360, 720)
	for gamma_c in dirs:
		gamma_c = np.radians(gamma_c)
		# current coefficient approx components
		Cx = c_x*np.cos(gamma_c)*np.absolute(np.cos(gamma_c))
		Cy = c_y*np.sin(gamma_c)*np.sin(gamma_c)
		y1.append(Cx)
		y2.append(Cy)

	plt.plot(dirs, y1, linewidth='2', label='C_c,x')
	plt.plot(dirs, y2, linewidth='2', label='C_c,y')
	plt.xlabel('Current direction [°]', fontsize='16')
	plt.ylabel('Coefficient value', fontsize='16')
	plt.title('Current coefficient approximation values', fontsize='16')
	plt.legend()
	plt.grid(True)
	plt.show()
