## Main
# Import classes
from Classes.Battery          import Battery

from Classes.ChargingStation  import ChargingStation as CS
from Classes.Ferry            import Ferry
from Classes.Grid             import Grid
from Classes.TransmissionLine import TransmissionLine as Line
from Classes.Model            import Model
from Classes.Output           import Output as Out
from DB                       import db
from ELCost                   import elcost as EL
import Classes.Cell as Cell
from External import pcu

# Import packages
import numpy  as np
import pandas as pd
import plotly
import plotly.graph_objs as go

from datetime import datetime  as dt
from datetime import timedelta as td

import matplotlib.pyplot as plt
from matplotlib import dates


# System definitions
TIMESTEP = 30 # [sec]
w_to_kW  = 10**(-3)

# Cost parameters
c_batt = 50  # Cell cost [NOK]
Gamma  = Cell.get_nominal_battery_life()


def main():
	# Construct output class
	out = Out()

	# Construct ferry model
	ferry            = Ferry(1800, 205920, None) #[capacity, #cells, input line]

	# Construct north harbour model
	cs_N             = CS(1, 1200, 137280, None, None)
	grid_N           = Grid(5000*10**3, None)
	line_grid_cs_N   = Line(grid_N, cs_N, 5000*10**3)
	line_cs_ferry_N  = Line(cs_N, ferry, 5000*10**3)

	# Construct south harbour model
	cs_S             = CS(2, 1200, 137280, None, None)
	grid_S           = Grid(5000*10**3, None)
	line_grid_cs_S   = Line(grid_S, cs_S, 5000*10**3)
	line_cs_ferry_S  = Line(cs_S, ferry, 5000*10**3)

	# Construct nightlay station model
	cs_N0			 = CS(0, 0*10**3, 0, None, None)
	grid_N0			 = Grid(5000*10**3, None)
	line_grid_cs_N0  = Line(grid_N0, cs_N0, 5000*10**3)
	line_cs_ferry_N0 = Line(cs_N0, ferry, 5000*10**3)

	# Update transmission line connections
	grid_N.line_output  = line_grid_cs_N
	cs_N.line_input     = line_grid_cs_N
	cs_N.line_output    = line_cs_ferry_N

	grid_S.line_output  = line_grid_cs_S
	cs_S.line_input     = line_grid_cs_S
	cs_S.line_output    = line_cs_ferry_S

	grid_N0.line_output = line_grid_cs_N0
	cs_N0.line_input    = line_grid_cs_N0
	cs_N0.line_output   = line_cs_ferry_N0

	ferry.line_input    = line_cs_ferry_N0

	# Pack model in structure
	model = Model()
	model.ferry = ferry
	model.grid  = [grid_N0, grid_N, grid_S]
	model.cs    = [cs_N0, cs_N, cs_S]

	# Query database data
	profile      = "Profile C"
	database     = "profileC.db"
	table        = "Profile"
	db_profile   = "C:\\Users\\noldu\\OneDrive\\NTNU\\Master thesis\\Application\\Profiles\\"+ profile + "\\"+database
	profile_data = db.query_all_db_data(db_profile, table)

	# Clear dash db
	model_db_file = 'C:\\Users\\noldu\\OneDrive\\NTNU\\Master thesis\\Application\\DB\\model.db'
	db.delete_table_records(model_db_file, 'Ferry')
	db.delete_table_records(model_db_file, 'cs_n')
	db.delete_table_records(model_db_file, 'cs_s')
	db.delete_table_records(model_db_file, 'el_cost')
	db.delete_table_records(model_db_file, 'ExternalData')

	## Simulate ferry operation ##
	index = 0
	termination_flag = False
	while termination_flag == False:

		# Get ferry mode
		model.ferry.mode = profile_data.at[index, 'mode']

		# Switch case for operational mode
		print(model.ferry.mode, profile_data.at[index, 'leg'], profile_data.at[index, 'datetime'])
		model = mode_switch(model.ferry.mode, profile_data, model, index)

		# Update DB & tables for use in Dash
		db.update_dash_db(model, profile_data, index)

		index = index + 1

		# Set weekend flags
		if index == 2312:
			EL.set_weekend_flag()
			pcu.set_weekend_flag()

		if index == 3030:
			out.ferry_Ah_working = (model.ferry.battery.cell.Ah_ctrl)	
			out.ferry_power_loss_working =  model.ferry.controller.C_loss
			out.ferry_aging_working = (c_batt/Gamma)*model.ferry.battery.cell.Ah_eff_ctrl*model.ferry.battery.NoC
			

		if index == 3083:
			out.csn_Ah_working = (model.cs[1].battery.cell.Ah_ctrl)
			out.csn_el_working = model.cs[1].controller.C_EL
			out.csn_power_loss_working =  model.cs[1].controller.C_loss
			out.csn_aging_working = (c_batt/Gamma)*model.cs[1].battery.cell.Ah_eff_ctrl*model.cs[1].battery.NoC
			out.csn_load_working = model.cs[1].controller.C_load

		if index == 3051:
			out.css_Ah_working = (model.cs[2].battery.cell.Ah_ctrl)
			out.css_el_working = model.cs[2].controller.C_EL
			out.css_power_loss_working =  model.cs[2].controller.C_loss
			out.css_aging_working = (c_batt/Gamma)*model.cs[2].battery.cell.Ah_eff_ctrl*model.cs[1].battery.NoC
			out.css_load_working = model.cs[2].controller.C_load


		# Termination criteria
		if index >= 5910:
			termination_flag = True


	## Results

	print('Simulation finished \n')
	NoC_ferry     = model.ferry.battery.NoC
	NoC_cs        = model.cs[1].battery.NoC

	NoC_ferry_Ah = 720
	NoC_cs_Ah    = 480

	total_Ah_eff       = model.ferry.battery.cell.Ah_eff*NoC_ferry_Ah + model.cs[1].battery.cell.Ah_eff*NoC_cs_Ah + model.cs[2].battery.cell.Ah_eff*NoC_cs_Ah
	total_Ah_ctrl      = model.ferry.battery.cell.Ah_ctrl*NoC_ferry_Ah + model.cs[1].battery.cell.Ah_ctrl*NoC_cs_Ah + model.cs[2].battery.cell.Ah_ctrl*NoC_cs_Ah
	total_Ah_eff_ctrl  = model.ferry.battery.cell.Ah_eff_ctrl*NoC_ferry_Ah + model.cs[1].battery.cell.Ah_eff_ctrl*NoC_cs_Ah + model.cs[2].battery.cell.Ah_eff_ctrl*NoC_cs_Ah

	total_energy       = model.cs[1].controller.total_energy + model.cs[2].controller.total_energy
	
	total_EL_cost      = model.cs[0].controller.C_EL + model.cs[1].controller.C_EL + model.cs[2].controller.C_EL
	total_EL_cost      = total_EL_cost + 18000*(1/365)*(1/24)*(1/(2*60))*index # Fixed tariff cost

	total_P_loss_cost  = model.ferry.controller.C_loss + model.cs[1].controller.C_loss + model.cs[2].controller.C_loss

	total_aging_cost   = (c_batt/Gamma)*model.ferry.battery.cell.Ah_eff_ctrl*NoC_ferry + (c_batt/Gamma)*model.cs[1].battery.cell.Ah_eff_ctrl*NoC_cs + (c_batt/Gamma)*model.cs[2].battery.cell.Ah_eff_ctrl*NoC_cs

	total_load_cost    = model.cs[1].controller.C_load + model.cs[2].controller.C_load

	total_cost         = total_EL_cost + total_P_loss_cost + total_aging_cost + total_load_cost

	print('Working day data: \n')
	print('Ah: ', out.css_Ah_working, out.csn_Ah_working, out.ferry_Ah_working)
	print('EL: ', out.css_el_working, out.csn_el_working)
	print('LOAD: ', out.css_load_working, out.csn_load_working)
	print('LOSS: ', out.css_power_loss_working, out.csn_power_loss_working, out.ferry_power_loss_working)
	print('AGING: ', out.ferry_aging_working, out.css_aging_working, out.csn_aging_working)
	print('\n')

	print('Total effective Ah: ', total_Ah_eff)
	print('Total controllable Ah: ', total_Ah_ctrl)
	print('Total effective controllable Ah: ', total_Ah_eff_ctrl)
	print('Total controllable energy [kWh]: ', total_energy,'\n')

	print('Price of electricity: ', total_EL_cost)
	print('Cost of power loss: ', total_P_loss_cost)
	print('Battery aging cost: ', total_aging_cost)
	print('Grid load cost: ', total_load_cost)
	print('Accumulated cost: ', total_cost, '\n')
	
	print('Ratio, El-cost/Ah: ', total_EL_cost/total_Ah_ctrl)
	print('Ratio, P_loss/Ah: ', total_P_loss_cost/total_Ah_ctrl)
	print('Ratio, Aging/Ah: ', total_aging_cost/total_Ah_ctrl)
	print('Ratio, P_load/Ah: ', total_load_cost/total_Ah_ctrl)
	print('Ratio, Cost/Ah: ', total_cost/total_Ah_ctrl)
	print('\n')


# Switch case for operational modes
def mode_switch(mode, profile, model, index):
	if mode == 'Nightlay':
		return nightlay_mode(profile, model, index)
	elif mode == 'Transit':
		return transit_mode(profile, model, index)
	elif mode == 'Conn':
		return conn_mode(profile, model, index)
	elif mode == 'Discon':
		return discon_mode(profile, model, index)
	elif mode == 'Docked':
		return docked_mode(profile, model, index)


# Mode functions
def nightlay_mode(profile, model, index):

	for cs in model.cs:
		if model.ferry.line_input == cs.line_output:
			# Ferry charging in nightlay
			cs.controller.set_ferry_chrg_pwr(cs, model, profile, index)
			cs.set_power(0, cs.controller.grid_power)
			cs.line_output.set_output_power(cs.grid_power)
			model.ferry.set_power(model.ferry.line_input.line_power)
			model.ferry.controller.set_power(model, profile, index)

			# Update ferry battery data
			model.ferry.battery.set_battery_power(model.ferry.controller.battery_power)
			model.ferry.battery.update_battery_SOC()
			model.ferry.battery.update_battery_SOH()
			model.ferry.battery.update_battery_terminal_voltage()
			model.ferry.battery.cell.update_controllable_Ah_throughput()
			model.ferry.battery.cell.update_effective_controllable_Ah_throughput()

			# Update charging station data
			cs.battery.set_battery_power(cs.controller.battery_power)
			cs.battery.update_battery_SOC()
			cs.battery.update_battery_SOH()
			cs.battery.update_battery_terminal_voltage()

			# Update EL cost
			time = dt.strptime(profile.at[index, 'datetime'], '%Y-%m-%d %H:%M:%S')
			cs.controller.update_EL_cost(time)

			# Update grid load cost
			cs.controller.update_load_cost(time)

			# Update power loss cost
			model.ferry.controller.update_power_loss_cost(model.ferry, time)

		else:
			# Charging stations not connected to ferry
			cs.controller.set_cs_chrg_pwr(cs, index)
			cs.set_power(cs.controller.battery_power, 0)

			# Update charging station data
			cs.battery.set_battery_power(cs.battery_power)
			cs.battery.update_battery_SOC()
			cs.battery.update_battery_SOH()
			cs.battery.update_battery_terminal_voltage()
			cs.battery.cell.update_controllable_Ah_throughput()
			cs.battery.cell.update_effective_controllable_Ah_throughput()

			# Update EL cost
			time = dt.strptime(profile.at[index, 'datetime'], '%Y-%m-%d %H:%M:%S')
			cs.controller.update_EL_cost(time)

			# Update grid load cost
			cs.controller.update_load_cost(time)

			# Update power loss cost
			cs.controller.update_power_loss_cost(cs, time)

	return model


def transit_mode(profile, model, index):
	# Supply total load demand
	model.ferry.controller.set_power(model, profile, index)
	model.ferry.battery.set_battery_power(model.ferry.controller.battery_power)

	# Update data in ferry battery
	model.ferry.battery.update_battery_SOC()
	model.ferry.battery.update_battery_SOH()
	model.ferry.battery.update_battery_terminal_voltage()

	# Set charging power and update data in charging station
	for cs in model.cs:
		cs.controller.set_cs_chrg_pwr(cs, index)
		cs.set_power(cs.controller.battery_power, 0)
		cs.battery.set_battery_power(cs.battery_power)
		cs.battery.update_battery_SOC()
		cs.battery.update_battery_SOH()
		cs.battery.update_battery_terminal_voltage()
		cs.battery.cell.update_controllable_Ah_throughput()
		cs.battery.cell.update_effective_controllable_Ah_throughput()

		# Update EL cost
		time = dt.strptime(profile.at[index, 'datetime'], '%Y-%m-%d %H:%M:%S')
		cs.controller.update_EL_cost(time)

		# Update grid load cost
		cs.controller.update_load_cost(time)

		# Update power loss cost
		cs.controller.update_power_loss_cost(cs, time)

	return model


def conn_mode(profile, model, index):

	model.ferry.controller.set_power(model, profile, index)
	model.ferry.battery.set_battery_power(model.ferry.controller.battery_power)

	# Update data in ferry battery
	model.ferry.battery.update_battery_SOC()
	model.ferry.battery.update_battery_SOH()
	model.ferry.battery.update_battery_terminal_voltage()

	## Connect ferry to actual charging station
	leg = profile.at[index, 'leg']
	if leg == 0:
		# Ferry at nightlay harbour
		model.ferry.line_input = model.cs[0].line_output
	elif leg == 1:
		# Ferry at south harbour
		model.ferry.line_input = model.cs[2].line_output
	else:
		# Ferry at north harbour
		model.ferry.line_input = model.cs[1].line_output
	

	for cs in model.cs:
		if model.ferry.line_input == cs.line_output:
			## Calculate ferry charging profile
			cs.controller.calculate_new_charging_profile(cs, model, profile, index)

		else:
			# Charging stations not connected to ferry
			cs.controller.set_cs_chrg_pwr(cs, index)
			cs.set_power(cs.controller.battery_power, 0)

			# Update charging station and battery data
			cs.battery.set_battery_power(cs.battery_power)
			cs.battery.update_battery_SOC()
			cs.battery.update_battery_SOH()
			cs.battery.update_battery_terminal_voltage()
			cs.battery.cell.update_controllable_Ah_throughput()
			cs.battery.cell.update_effective_controllable_Ah_throughput()

			# Update EL cost
			time = dt.strptime(profile.at[index, 'datetime'], '%Y-%m-%d %H:%M:%S')
			cs.controller.update_EL_cost(time)

			# Update grid load cost
			cs.controller.update_load_cost(time)

			# Update power loss cost
			cs.controller.update_power_loss_cost(cs, time)

	## return model
	return model


def discon_mode(profile, model, index):

	model.ferry.controller.set_power(model, profile, index)
	model.ferry.battery.set_battery_power(model.ferry.controller.battery_power)

	# Update data in ferry battery
	model.ferry.battery.update_battery_SOC()
	model.ferry.battery.update_battery_SOH()
	model.ferry.battery.update_battery_terminal_voltage()

	for cs in model.cs:
		if cs.line_output == model.ferry.line_input:
			# Charging station connected to ferry
			cs.controller.battery_power = 0
			cs.controller.ferry_grid_power = 0
			cs.battery.set_battery_power(0)

			# Calculate recharging profile for charging station
			cs.controller.calculate_new_charging_profile(cs, model, profile, index)
		else:
			# Charging station on other harbour[s]
			cs.controller.set_cs_chrg_pwr(cs, index)
			cs.set_power(cs.controller.battery_power, 0)
			cs.battery.set_battery_power(cs.battery_power)
			cs.battery.update_battery_SOC()
			cs.battery.update_battery_SOH()
			cs.battery.update_battery_terminal_voltage()
			cs.battery.cell.update_controllable_Ah_throughput()
			cs.battery.cell.update_effective_controllable_Ah_throughput()

			# Update EL cost
			time = dt.strptime(profile.at[index, 'datetime'], '%Y-%m-%d %H:%M:%S')
			cs.controller.update_EL_cost(time)

			# Update grid load cost
			cs.controller.update_load_cost(time)

			# Update power loss cost
			cs.controller.update_power_loss_cost(cs, time)

	# Disconnect ferry from charging station
	model.ferry.line_input = None

	return model


def docked_mode(profile, model, index):

	for cs in model.cs:
		if model.ferry.line_input == cs.line_output:
			# Ferry charging in nightlay
			cs.controller.set_ferry_chrg_pwr(cs, model, profile, index)
			cs.set_power(np.absolute(cs.controller.battery_power), cs.controller.grid_power)
			cs.line_output.set_output_power(cs.grid_power + cs.battery_power)
			model.ferry.set_power(model.ferry.line_input.line_power)
			model.ferry.controller.set_power(model, profile, index)

			## Update ferry battery data
			model.ferry.battery.set_battery_power(model.ferry.controller.battery_power)
			model.ferry.battery.update_battery_SOC()
			model.ferry.battery.update_battery_SOH()
			model.ferry.battery.update_battery_terminal_voltage()
			model.ferry.battery.cell.update_controllable_Ah_throughput()
			model.ferry.battery.cell.update_effective_controllable_Ah_throughput()

			# Update power loss cost
			model.ferry.controller.update_power_loss_cost(model.ferry, dt.strptime(profile.at[index, 'datetime'], '%Y-%m-%d %H:%M:%S'))

			## Update charging station data
			cs.battery.set_battery_power(cs.controller.battery_power)
			cs.battery.update_battery_SOC()
			cs.battery.update_battery_SOH()
			cs.battery.update_battery_terminal_voltage()
			cs.battery.cell.update_controllable_Ah_throughput()
			cs.battery.cell.update_effective_controllable_Ah_throughput()

			# Update EL cost
			time = dt.strptime(profile.at[index, 'datetime'], '%Y-%m-%d %H:%M:%S')
			cs.controller.update_EL_cost(time)

			# Update grid load cost
			cs.controller.update_load_cost(time)

			# Update power loss cost
			cs.controller.update_power_loss_cost(cs, time)

		else:
			# Charging stations not connected to ferry
			cs.controller.set_cs_chrg_pwr(cs, index)
			cs.set_power(cs.controller.battery_power, 0)

			# Update charging station data
			cs.battery.set_battery_power(cs.battery_power)
			cs.battery.update_battery_SOC()
			cs.battery.update_battery_SOH()
			cs.battery.update_battery_terminal_voltage()
			cs.battery.cell.update_controllable_Ah_throughput()
			cs.battery.cell.update_effective_controllable_Ah_throughput()

			# Update EL cost
			time = dt.strptime(profile.at[index, 'datetime'], '%Y-%m-%d %H:%M:%S')
			cs.controller.update_EL_cost(time)

			# Update grid load cost
			cs.controller.update_load_cost(time)

			# Update power loss cost
			cs.controller.update_power_loss_cost(cs, time)

	return model

def plot_op_profile(out):
	fig1, axs = plt.subplots(2,1)
	# date_list = dates.date2num(out.datetime)
	axs[0].plot(out.datetime, out.op_working, linewidth='2')
	#axs[0].set_xlabel('Date and hour', fontsize='16')
	axs[0].set_ylabel('Ferry power [kW]', fontsize='16')
	axs[0].set_title('Operational profile - weekend day', fontsize='16')
	axs[0].grid(True)

	dt_scope  = []
	pwr_scope = []
	for i in range(1555, 1680):
		dt_scope.append(out.datetime[i])
		pwr_scope.append(out.op_working[i])

	axs[1].plot(dt_scope, pwr_scope, linewidth='2')
	#plt.gcf().autofmt_xdate()
	axs[1].set_xlabel('Date and time', fontsize='16')
	axs[1].set_ylabel('Ferry power [kW]', fontsize='16')
	axs[1].grid(True)
	plt.show()


def plot_Ah(out):
	plt.subplot(311)
	plt.plot(out.ferry_Ah)
	plt.grid(True)

	plt.subplot(312)
	plt.plot(out.css_Ah)
	plt.grid(True)

	plt.subplot(313)
	plt.plot(out.csn_Ah)
	plt.grid(True)

	plt.show()


def generate_plots(out):
	# ferry SOC
	plotly.offline.plot({
	    "data": [go.Scatter(x=out.datetime, y=out.ferry_SOC, line=dict(shape='hv'))],
	    "layout": go.Layout(title="Ferry SOC")
	}, auto_open=True)


	# Ferry power
	plotly.offline.plot({
	    "data": [go.Scatter(x=out.datetime, y=out.ferry_bat_pwr, line=dict(shape='hv'))],
	    "layout": go.Layout(title="Ferry pwr")
	}, auto_open=True)


	# North CS SOC
	plotly.offline.plot({
	    "data": [go.Scatter(x=out.datetime, y=out.cs_N_SOC, line=dict(shape='hv'))],
	    "layout": go.Layout(title="North CS SOC")
	}, auto_open=True)


	# North CS battery power
	plotly.offline.plot({
	    "data": [go.Scatter(x=out.datetime, y=out.cs_N_pwr, line=dict(shape='hv'))],
	    "layout": go.Layout(title="North CS battery power")
	}, auto_open=True)



# RUN PROGRAM
if __name__ == '__main__': main()